@@ -1,6 +1,6 @@
 /*
- * Copyright 2007-2008 Secure Endpoints Inc.  
- * 
+ * Copyright 2007-2008 Secure Endpoints Inc.
+ *
  * All Rights Reserved.
  *
  * This software has been released under the terms of the IBM Public
@@ -13,8 +13,8 @@
 #include <windows.h>
 #include <stdio.h>
 #include <stdlib.h>
-#include <assert.h>
 #include "afsd.h"
+#include <assert.h>
 #include <strsafe.h>
 
 #ifdef USE_BPLUS
@@ -128,10 +128,10 @@ static DWORD TlsDataIndex;
 
 long cm_InitBPlusDir(void)
 {
-    if ((TlsKeyIndex = TlsAlloc()) == TLS_OUT_OF_INDEXES) 
+    if ((TlsKeyIndex = TlsAlloc()) == TLS_OUT_OF_INDEXES)
         return 0;
 
-    if ((TlsDataIndex = TlsAlloc()) == TLS_OUT_OF_INDEXES) 
+    if ((TlsDataIndex = TlsAlloc()) == TLS_OUT_OF_INDEXES)
         return 0;
 
     return 1;
@@ -193,10 +193,10 @@ void freeBtree(Tree *B)
 /* access key and data values for B+tree methods */
 /* pass values to getSlot(), descend...() */
 static keyT getfunkey(Tree *B) {
-    keyT *tlsKey; 
- 
-    // Retrieve a data pointer for the current thread. 
-    tlsKey = (keyT *) TlsGetValue(TlsKeyIndex); 
+    keyT *tlsKey;
+
+    // Retrieve a data pointer for the current thread.
+    tlsKey = (keyT *) TlsGetValue(TlsKeyIndex);
     if (tlsKey == NULL) {
         if (GetLastError() != ERROR_SUCCESS)
             osi_panic("TlsGetValue failed", __FILE__, __LINE__);
@@ -208,10 +208,10 @@ static keyT getfunkey(Tree *B) {
 }
 
 static dataT getfundata(Tree *B) {
-    dataT *tlsData; 
- 
-    // Retrieve a data pointer for the current thread. 
-    tlsData = (dataT *) TlsGetValue(TlsDataIndex); 
+    dataT *tlsData;
+
+    // Retrieve a data pointer for the current thread.
+    tlsData = (dataT *) TlsGetValue(TlsDataIndex);
     if (tlsData == NULL) {
         if (GetLastError() != ERROR_SUCCESS)
             osi_panic("TlsGetValue failed", __FILE__, __LINE__);
@@ -225,14 +225,14 @@ static dataT getfundata(Tree *B) {
 static void setfunkey(Tree *B, keyT theKey) {
     keyT *tlsKey;
 
-    tlsKey = (keyT *) TlsGetValue(TlsKeyIndex); 
+    tlsKey = (keyT *) TlsGetValue(TlsKeyIndex);
     if (tlsKey == NULL) {
         if (GetLastError() != ERROR_SUCCESS)
             osi_panic("TlsGetValue failed", __FILE__, __LINE__);
 
         tlsKey = malloc(sizeof(keyT));
-        
-        if (!TlsSetValue(TlsKeyIndex, tlsKey)) 
+
+        if (!TlsSetValue(TlsKeyIndex, tlsKey))
             osi_panic("TlsSetValue failed", __FILE__, __LINE__);
     }
 
@@ -242,14 +242,14 @@ static void setfunkey(Tree *B, keyT theKey) {
 static void setfundata(Tree *B, dataT theData) {
     dataT *tlsData;
 
-    tlsData = (dataT *) TlsGetValue(TlsDataIndex); 
+    tlsData = (dataT *) TlsGetValue(TlsDataIndex);
     if (tlsData == NULL) {
         if (GetLastError() != ERROR_SUCCESS)
             osi_panic("TlsGetValue failed", __FILE__, __LINE__);
 
         tlsData = malloc(sizeof(dataT));
-        
-        if (!TlsSetValue(TlsDataIndex, tlsData)) 
+
+        if (!TlsSetValue(TlsDataIndex, tlsData))
             osi_panic("TlsSetValue failed", __FILE__, __LINE__);
     }
 
@@ -289,11 +289,11 @@ Nptr bplus_Lookup(Tree *B, keyT key)
 
         StringCbPrintfA(B->message, sizeof(B->message), "LOOKUP: %s found on page %d value (%d.%d.%d).\n",
                  key.name,
-                 getnodenumber(B, leafNode), 
-                 data.fid.volume, 
+                 getnodenumber(B, leafNode),
+                 data.fid.volume,
                  data.fid.vnode,
                  data.fid.unique);
-    } else 
+    } else
         StringCbPrintfA(B->message, sizeof(B->message), "LOOKUP: not found!\n");
     OutputDebugString(B->message);
 #endif
@@ -356,7 +356,7 @@ static int findKey(Tree *B, Nptr curr, int lo, int hi)
 
 #ifdef DEBUG_BTREE
         if (findslot == BTERROR) {
-            StringCbPrintfA(B->message, sizeof(B->message), "FINDKEY: (lo %d hi %d) Bad key ordering on node %d (0x%p)\n", 
+            StringCbPrintfA(B->message, sizeof(B->message), "FINDKEY: (lo %d hi %d) Bad key ordering on node %d (0x%p)\n",
                     lo, hi, getnodenumber(B, curr), curr);
             osi_Log1(afsd_logp, "BPlus: %s", osi_LogSaveString(afsd_logp, B->message));
         }
@@ -373,7 +373,7 @@ static int findKey(Tree *B, Nptr curr, int lo, int hi)
                 findslot = findKey(B, curr, mid + 1, hi);
             break;
         case BTERROR:
-            StringCbPrintfA(B->message, sizeof(B->message), "FINDKEY: (lo %d hi %d) Bad key ordering on node %d (0x%p)\n", 
+            StringCbPrintfA(B->message, sizeof(B->message), "FINDKEY: (lo %d hi %d) Bad key ordering on node %d (0x%p)\n",
                     lo, hi, getnodenumber(B, curr), curr);
             osi_Log1(afsd_logp, "BPlus: %s", osi_LogSaveString(afsd_logp, B->message));
         }
@@ -381,7 +381,7 @@ static int findKey(Tree *B, Nptr curr, int lo, int hi)
 
     if (isleaf(curr) && findslot == 0)
     {
-        StringCbPrintfA(B->message, sizeof(B->message), "FINDKEY: (lo %d hi %d) findslot %d is invalid for leaf nodes, bad key ordering on node %d (0x%p)\n", 
+        StringCbPrintfA(B->message, sizeof(B->message), "FINDKEY: (lo %d hi %d) findslot %d is invalid for leaf nodes, bad key ordering on node %d (0x%p)\n",
                 lo, hi, findslot, getnodenumber(B, curr), curr);
         osi_Log1(afsd_logp, "BPlus: %s", osi_LogSaveString(afsd_logp, B->message));
     }
@@ -402,8 +402,8 @@ static int bestMatch(Tree *B, Nptr curr, int slot)
             if (isleaf(curr))
                  findslot = BTLOWER;	/* not found in the tree */
             else
-                 findslot = 0;				
-        } 
+                 findslot = 0;
+        }
         else if ((comp = comparekeys(B)(getfunkey(B), getkey(curr, slot - 1), 0)) >= 0) {
             findslot = slot - 1;
         } else if (comp < diff) {
@@ -416,9 +416,9 @@ static int bestMatch(Tree *B, Nptr curr, int slot)
         }
     } else {			/* or check following slot */
         if (slot == numentries(curr)) {
-            if (isleaf(curr) && numentries(curr) == getfanout(B)) 
+            if (isleaf(curr) && numentries(curr) == getfanout(B))
                 findslot = BTUPPER;
-            else 
+            else
                 findslot = slot;
         } else if ((comp = comparekeys(B)(getfunkey(B), getkey(curr, slot + 1), 0)) < 0) {
             findslot = slot;
@@ -436,7 +436,7 @@ static int bestMatch(Tree *B, Nptr curr, int slot)
 
     if (findslot == BTERROR || isleaf(curr) && findslot == 0)
     {
-        StringCbPrintfA(B->message, sizeof(B->message), "BESTMATCH: node %d (0x%p) slot %d diff %d comp %d findslot %d\n", 
+        StringCbPrintfA(B->message, sizeof(B->message), "BESTMATCH: node %d (0x%p) slot %d diff %d comp %d findslot %d\n",
                 getnodenumber(B, curr), curr, slot, diff, comp, findslot);
         osi_Log1(afsd_logp, "BPlus: %s", osi_LogSaveString(afsd_logp, B->message));
     }
@@ -469,7 +469,7 @@ void insert(Tree *B, keyT key, dataT data)
 
 
 /*****************   recurse down and split back up   ******************/
-static Nptr 
+static Nptr
 descendSplit(Tree *B, Nptr curr)
 {
     Nptr	downNode = NONODE, sibling = NONODE;
@@ -522,7 +522,7 @@ descendSplit(Tree *B, Nptr curr)
 }
 
 /***************   determine location of inserted key   ****************/
-static void 
+static void
 insertEntry(Tree *B, Nptr currNode, int slot, Nptr sibling, Nptr downPtr)
 {
     int split, i, j, k, x, y;
@@ -552,7 +552,7 @@ insertEntry(Tree *B, Nptr currNode, int slot, Nptr sibling, Nptr downPtr)
         k = (slot >= split ? 1 : 0);
 
         /*
-         * Move entries from the top half of the current node to 
+         * Move entries from the top half of the current node to
          * to the sibling node.
          * The number of entries to move is dependent upon where
          * the new entry is going to be added in relationship to
@@ -563,7 +563,7 @@ insertEntry(Tree *B, Nptr currNode, int slot, Nptr sibling, Nptr downPtr)
          *
          * If the node that is being split is an internal node (i != 0)
          * then we move one less entry due to the extra down pointer
-         * when the split slot is not equal to the insertion slot 
+         * when the split slot is not equal to the insertion slot
          */
         for (x = split + k + j * i, y = 1; x <= getfanout(B); x++, y++) {
             xferentry(currNode, x, sibling, y);	/* copy entries to sibling */
@@ -611,7 +611,7 @@ insertEntry(Tree *B, Nptr currNode, int slot, Nptr sibling, Nptr downPtr)
 
             /* set key separating nodes */
             if (isleaf(sibling))
-                key = getkey(sibling, 1); 
+                key = getkey(sibling, 1);
             else {
                 Nptr leaf = getfirstnode(sibling);
                 while ( isinternal(leaf) )
@@ -626,7 +626,7 @@ insertEntry(Tree *B, Nptr currNode, int slot, Nptr sibling, Nptr downPtr)
 }
 
 /************   place key into appropriate node & slot   ***************/
-static void 
+static void
 placeEntry(Tree *B, Nptr node, int slot, Nptr downPtr)
 {
     int x;
@@ -660,7 +660,7 @@ placeEntry(Tree *B, Nptr node, int slot, Nptr downPtr)
 
 
 /*****************   split full node and set flags   *******************/
-static Nptr 
+static Nptr
 split(Tree *B, Nptr node)
 {
     Nptr sibling;
@@ -751,7 +751,7 @@ void delete(Tree *B, keyT key)
 
 
 /**********************   remove old root node   ***********************/
-static void 
+static void
 collapseRoot(Tree *B, Nptr oldRoot, Nptr newRoot)
 {
 
@@ -771,7 +771,7 @@ collapseRoot(Tree *B, Nptr oldRoot, Nptr newRoot)
 
 
 /****************   recurse down and balance back up   *****************/
-static Nptr 
+static Nptr
 descendBalance(Tree *B, Nptr curr, Nptr left, Nptr right, Nptr lAnc, Nptr rAnc, Nptr parent)
 {
     Nptr	newMe=NONODE, myLeft=NONODE, myRight=NONODE, lAnchor=NONODE, rAnchor=NONODE, newNode=NONODE;
@@ -830,7 +830,7 @@ descendBalance(Tree *B, Nptr curr, Nptr left, Nptr right, Nptr lAnc, Nptr rAnc,
         }
         newMe = descendBalance(B, newNode, myLeft, myRight, lAnchor, rAnchor, curr);
     }
-    else if ((slot > 0) && !comparekeys(B)(getfunkey(B), getkey(curr, slot), 0)) 
+    else if ((slot > 0) && !comparekeys(B)(getfunkey(B), getkey(curr, slot), 0))
     {
         Nptr        next;
         int         exact = 0;
@@ -840,7 +840,7 @@ descendBalance(Tree *B, Nptr curr, Nptr left, Nptr right, Nptr lAnc, Nptr rAnc,
         next = getdatanext(newNode);
 
         /*
-         * We only delete exact matches.  
+         * We only delete exact matches.
          */
         if (!comparekeys(B)(getfunkey(B), getdatakey(newNode), EXACT_MATCH)) {
             /* exact match, free the first entry */
@@ -849,7 +849,7 @@ descendBalance(Tree *B, Nptr curr, Nptr left, Nptr right, Nptr lAnc, Nptr rAnc,
             if (next == NONODE) {
                 /* delete this key as there are no more data values */
                 newMe = newNode;
-            } else { 
+            } else {
                 /* otherwise, there are more and we leave the key in place */
                 setnode(curr, slot, next);
                 putFreeNode(B, newNode);
@@ -859,14 +859,14 @@ descendBalance(Tree *B, Nptr curr, Nptr left, Nptr right, Nptr lAnc, Nptr rAnc,
                 setmergepath(B, NONODE);
             }
         } else if (next == NONODE) {
-            /* 
+            /*
              * we didn't find an exact match and there are no more
              * choices.  so we leave it alone and remove nothing.
              */
             newMe = NONODE;
             setmergepath(B, NONODE);
         } else {
-            /* The first data node doesn't match but there are other 
+            /* The first data node doesn't match but there are other
              * options.  So we must determine if any of the next nodes
              * are the one we are looking for.
              */
@@ -882,13 +882,13 @@ descendBalance(Tree *B, Nptr curr, Nptr left, Nptr right, Nptr lAnc, Nptr rAnc,
                 prev = next;
                 next = getdatanext(next);
             }
-            
+
             /* do not delete the key */
             newMe = NONODE;
             setmergepath(B, NONODE);
         }
     }
-    else 
+    else
     {
         newMe = NONODE;		/* no deletion possible, key not found */
         setmergepath(B, NONODE);
@@ -999,7 +999,7 @@ removeEntry(Tree *B, Nptr curr, int slot)
 
 
 /*******   merge a node pair & set emptied node up for removal   *******/
-static Nptr 
+static Nptr
 merge(Tree *B, Nptr left, Nptr right, Nptr anchor)
 {
     int	x, y, z;
@@ -1056,15 +1056,15 @@ merge(Tree *B, Nptr left, Nptr right, Nptr anchor)
 
 
 /******   shift entries in a node pair & adjust anchor key value   *****/
-static Nptr 
+static Nptr
 shift(Tree *B, Nptr left, Nptr right, Nptr anchor)
 {
     int	i, x, y, z;
 
 #ifdef DEBUG_BTREE
-    StringCbPrintfA(B->message, sizeof(B->message), "SHIFT:  left %d, right %d, anchor %d.\n", 
-             getnodenumber(B, left), 
-             getnodenumber(B, right), 
+    StringCbPrintfA(B->message, sizeof(B->message), "SHIFT:  left %d, right %d, anchor %d.\n",
+             getnodenumber(B, left),
+             getnodenumber(B, right),
              getnodenumber(B, anchor));
     osi_Log1(afsd_logp, "BPlus: %s", osi_LogSaveString(afsd_logp, B->message));
     showNode(B, "pre-shift anchor", anchor);
@@ -1073,7 +1073,7 @@ shift(Tree *B, Nptr left, Nptr right, Nptr anchor)
 #endif
 
     i = isinternal(left);
-    
+
     if (numentries(left) < numentries(right)) {	/* shift entries to left */
         y = (numentries(right) - numentries(left)) >> 1;
         x = numentries(left) + y;
@@ -1116,7 +1116,7 @@ shift(Tree *B, Nptr left, Nptr right, Nptr anchor)
 
         for (z = numentries(right); z > 0; z--)	/* adjust increased node */
             pushentry(right, z, y);
-        
+
         setfunkey(B, getkey(left, x));			/* set new anchor key value */
         z = getSlot(B, anchor);
         if (z <= BTERROR)
@@ -1145,7 +1145,7 @@ shift(Tree *B, Nptr left, Nptr right, Nptr anchor)
             xferentry(left, x, right, y);		/* transfer entries over */
             clrentry(left, x);
         }
-    } 
+    }
 #ifdef DEBUG_BTREE
     else {
         DebugBreak();
@@ -1185,7 +1185,7 @@ _clrentry(Nptr node, int entry)
 }
 
 static void
-_pushentry(Nptr node, int entry, int offset) 
+_pushentry(Nptr node, int entry, int offset)
 {
     if (getkey(node,entry + offset).name != NULL)
         free(getkey(node,entry + offset).name);
@@ -1246,7 +1246,7 @@ _setentry(Nptr node, int entry, keyT key, Nptr downNode)
 \***********************************************************************/
 
 /*********************   Set up initial pool of free nodes   ***********/
-static void 
+static void
 initFreeNodePool(Tree *B, int quantity)
 {
     int	i;
@@ -1271,7 +1271,7 @@ initFreeNodePool(Tree *B, int quantity)
     }
     setnextnode(p, NONODE);		/* indicates end of free node list */
     setallnode(p, NONODE);              /* indicates end of all node list */
-    
+
     setpoolsize(B, quantity);
 }
 
@@ -1310,7 +1310,7 @@ cleanupNodePool(Tree *B)
 }
 
 /**************   take a free B+tree node from the pool   **************/
-static Nptr 
+static Nptr
 getFreeNode(Tree *B)
 {
     Nptr newNode = getfirstfreenode(B);
@@ -1335,7 +1335,7 @@ getFreeNode(Tree *B)
 
 
 /*************   return a deleted B+tree node to the pool   ************/
-static void 
+static void
 putFreeNode(Tree *B, Nptr node)
 {
     int i;
@@ -1365,7 +1365,7 @@ putFreeNode(Tree *B, Nptr node)
 
 
 /*******   fill a free data node with a key and associated data   ******/
-static Nptr 
+static Nptr
 getDataNode(Tree *B, keyT key, dataT data)
 {
     Nptr	newNode = getFreeNode(B);
@@ -1447,7 +1447,7 @@ void showBtree(Tree *B)
     osi_Log1(afsd_logp, "BPlus: %s", osi_LogSaveString(afsd_logp, B->message));
 }
 
-void 
+void
 listBtreeNodes(Tree *B, const char * parent_desc, Nptr node)
 {
     int i;
@@ -1458,17 +1458,17 @@ listBtreeNodes(Tree *B, const char * parent_desc, Nptr node)
         StringCbPrintfA(B->message, sizeof(B->message), "%s - NoNode!!!\n");
         osi_Log1(afsd_logp, "BPlus: %s", osi_LogSaveString(afsd_logp, B->message));
         return;
-    } 
-    
+    }
+
     if (!isnode(node))
     {
         data = getdatavalue(node);
-        StringCbPrintfA(B->message, sizeof(B->message), "%s - data node %d (%d.%d.%d)\n", 
+        StringCbPrintfA(B->message, sizeof(B->message), "%s - data node %d (%d.%d.%d)\n",
                  parent_desc, getnodenumber(B, node),
                  data.fid.volume, data.fid.vnode, data.fid.unique);
         osi_Log1(afsd_logp, "BPlus: %s", osi_LogSaveString(afsd_logp, B->message));
         return;
-    } else 
+    } else
         showNode(B, parent_desc, node);
 
     if ( isinternal(node) || isroot(node) ) {
@@ -1482,7 +1482,7 @@ listBtreeNodes(Tree *B, const char * parent_desc, Nptr node)
 }
 
 /***********************   B+tree data printer   ***********************/
-void 
+void
 listBtreeValues(Tree *B, Nptr n, int num)
 {
     int slot;
@@ -1497,7 +1497,7 @@ listBtreeValues(Tree *B, Nptr n, int num)
         }
         prev = getkey(n, slot);
         data = getdatavalue(getnode(n, slot));
-        StringCbPrintfA(B->message, sizeof(B->message), "%8S (%d.%d.%d)\n", 
+        StringCbPrintfA(B->message, sizeof(B->message), "%8S (%d.%d.%d)\n",
                 prev.name, data.fid.volume, data.fid.vnode, data.fid.unique);
         osi_Log1(afsd_logp, "BPlus: %s", osi_LogSaveString(afsd_logp, B->message));
         if (++slot > numentries(n))
@@ -1508,14 +1508,14 @@ listBtreeValues(Tree *B, Nptr n, int num)
 }
 
 /********************   entire B+tree data printer   *******************/
-void 
+void
 listAllBtreeValues(Tree *B)
 {
     listBtreeValues(B, getleaf(B), BTERROR);
 }
 #endif /* DEBUG_BTREE */
 
-void 
+void
 findAllBtreeValues(Tree *B)
 {
     int num = -1;
@@ -1536,7 +1536,7 @@ findAllBtreeValues(Tree *B)
         if ( l != n ){
             if (l == NONODE)
                 StringCbPrintfA(B->message, sizeof(B->message),"BOMB %8S cannot be found\n", prev.name);
-            else 
+            else
                 StringCbPrintfA(B->message, sizeof(B->message),"BOMB lookup(%8S) finds wrong node\n", prev.name);
             osi_Log1(afsd_logp, "BPlus: %s", osi_LogSaveString(afsd_logp, B->message));
 #ifdef DEBUG_BTREE
@@ -1546,17 +1546,17 @@ findAllBtreeValues(Tree *B)
 
         if (++slot > numentries(n))
             n = getnextnode(n), slot = 1;
-    }   
+    }
 }
 
-/* 
+/*
  * the return must be -1, 0, or 1.  stricmp() in MSVC 8.0
  * does not return only those values.
  *
  * the sorting of the tree is by case insensitive sort order
  * therefore, unless the strings actually match via a case
  * insensitive search do we want to perform the case sensitive
- * match.  Otherwise, the search order might be considered 
+ * match.  Otherwise, the search order might be considered
  * to be inconsistent when the EXACT_MATCH flag is set.
  */
 int
@@ -1581,13 +1581,17 @@ cm_BPlusDirLookupOriginalName(cm_dirOp_t * op, clientchar_t *centry,
     fschar_t * fsname = NULL;
     normchar_t * entry = NULL;
 
-    if (op->scp->dirBplus == NULL || 
-        op->dataVersion != op->scp->dirDataVersion) {
+    if (op->scp->dirBplus == NULL ||
+        op->dataVersion > op->scp->dirDataVersion) {
         rc = EINVAL;
         goto done;
     }
 
     entry = cm_ClientStringToNormStringAlloc(centry, -1, NULL);
+    if (!entry) {
+        rc = EINVAL;
+        goto done;
+    }
     key.name = entry;
 
     lock_AssertAny(&op->scp->dirlock);
@@ -1609,7 +1613,7 @@ cm_BPlusDirLookupOriginalName(cm_dirOp_t * op, clientchar_t *centry,
          */
         slot = getSlot(op->scp->dirBplus, leafNode);
         if (slot <= BTERROR) {
-            op->scp->dirDataVersion = 0;
+            op->scp->dirDataVersion = CM_SCACHE_VERSION_BAD;
             rc = (slot == BTERROR ? EINVAL : ENOENT);
             goto done;
         }
@@ -1635,7 +1639,7 @@ cm_BPlusDirLookupOriginalName(cm_dirOp_t * op, clientchar_t *centry,
         } else {
             rc = CM_ERROR_AMBIGUOUS_FILENAME;
             bplus_lookup_ambiguous++;
-        } 
+        }
     } else {
         rc = ENOENT;
         bplus_lookup_misses++;
@@ -1673,13 +1677,17 @@ cm_BPlusDirLookup(cm_dirOp_t * op, clientchar_t * centry, cm_fid_t * cfid)
     Nptr leafNode = NONODE;
     LARGE_INTEGER start, end;
 
-    if (op->scp->dirBplus == NULL || 
-        op->dataVersion != op->scp->dirDataVersion) {
+    if (op->scp->dirBplus == NULL ||
+        op->dataVersion > op->scp->dirDataVersion) {
         rc = EINVAL;
         goto done;
     }
 
     entry = cm_ClientStringToNormStringAlloc(centry, -1, NULL);
+    if (!entry) {
+        rc = EINVAL;
+        goto done;
+    }
     key.name = entry;
 
     lock_AssertAny(&op->scp->dirlock);
@@ -1727,7 +1735,7 @@ cm_BPlusDirLookup(cm_dirOp_t * op, clientchar_t * centry, cm_fid_t * cfid)
         } else {
             rc = CM_ERROR_AMBIGUOUS_FILENAME;
             bplus_lookup_ambiguous++;
-        } 
+        }
     } else {
             rc = ENOENT;
         bplus_lookup_misses++;
@@ -1745,7 +1753,7 @@ cm_BPlusDirLookup(cm_dirOp_t * op, clientchar_t * centry, cm_fid_t * cfid)
 }
 
 
-/* 
+/*
    On entry:
        op->scp->dirlock is write locked
 
@@ -1760,13 +1768,17 @@ long cm_BPlusDirCreateEntry(cm_dirOp_t * op, clientchar_t * entry, cm_fid_t * cf
     LARGE_INTEGER start, end;
     normchar_t * normalizedName = NULL;
 
-    if (op->scp->dirBplus == NULL || 
+    if (op->scp->dirBplus == NULL ||
         op->dataVersion != op->scp->dirDataVersion) {
         rc = EINVAL;
         goto done;
     }
 
     normalizedName = cm_ClientStringToNormStringAlloc(entry, -1, NULL);
+    if (!normalizedName) {
+        rc = EINVAL;
+        goto done;
+    }
     key.name = normalizedName;
 
     lock_AssertWrite(&op->scp->dirlock);
@@ -1780,7 +1792,7 @@ long cm_BPlusDirCreateEntry(cm_dirOp_t * op, clientchar_t * entry, cm_fid_t * cf
     bplus_create_entry++;
 
     insert(op->scp->dirBplus, key, data);
-    
+
     if (!cm_Is8Dot3(entry)) {
         cm_dirFid_t dfid;
         clientchar_t wshortName[13];
@@ -1811,7 +1823,7 @@ long cm_BPlusDirCreateEntry(cm_dirOp_t * op, clientchar_t * entry, cm_fid_t * cf
     return rc;
 }
 
-/* 
+/*
    On entry:
        op->scp->dirlock is write locked
 
@@ -1826,13 +1838,17 @@ int  cm_BPlusDirDeleteEntry(cm_dirOp_t * op, clientchar_t *centry)
     LARGE_INTEGER start, end;
     normchar_t * normalizedEntry = NULL;
 
-    if (op->scp->dirBplus == NULL || 
+    if (op->scp->dirBplus == NULL ||
         op->dataVersion != op->scp->dirDataVersion) {
         rc = EINVAL;
         goto done;
     }
 
     normalizedEntry = cm_ClientStringToNormStringAlloc(centry, -1, NULL);
+    if (!normalizedEntry) {
+        rc = EINVAL;
+        goto done;
+    }
     key.name = normalizedEntry;
 
     lock_AssertWrite(&op->scp->dirlock);
@@ -1885,24 +1901,23 @@ int  cm_BPlusDirDeleteEntry(cm_dirOp_t * op, clientchar_t *centry)
                     rc = CM_ERROR_INEXACT_MATCH;
                 } else {
                     rc = CM_ERROR_AMBIGUOUS_FILENAME;
-                } 
-            }
+                }
 
+                if (rc != CM_ERROR_AMBIGUOUS_FILENAME) {
+                    dfid.vnode = htonl(fid.vnode);
+                    dfid.unique = htonl(fid.unique);
+                    cm_Gen8Dot3NameIntW(centry, &dfid, shortName, NULL);
 
-            if (rc != CM_ERROR_AMBIGUOUS_FILENAME) {
-                dfid.vnode = htonl(fid.vnode);
-                dfid.unique = htonl(fid.unique);
-                cm_Gen8Dot3NameIntW(centry, &dfid, shortName, NULL);
-
-                /* delete first the long name and then the short name */
-                delete(op->scp->dirBplus, key);
-                key.name = shortName;
-                delete(op->scp->dirBplus, key);
-            }
+                    /* delete first the long name and then the short name */
+                    delete(op->scp->dirBplus, key);
+                    key.name = shortName;
+                    delete(op->scp->dirBplus, key);
+                }
+            } /* !NONODE */
         } else {
             clientchar_t * cname = NULL;
 
-            /* We need to lookup the 8dot3 name to determine what the 
+            /* We need to lookup the 8dot3 name to determine what the
              * matching long name is
              */
             leafNode = bplus_Lookup(op->scp->dirBplus, key);
@@ -1944,7 +1959,7 @@ int  cm_BPlusDirDeleteEntry(cm_dirOp_t * op, clientchar_t *centry)
                     rc = CM_ERROR_INEXACT_MATCH;
                 } else {
                     rc = CM_ERROR_AMBIGUOUS_FILENAME;
-                } 
+                }
             }
 
             if (rc != CM_ERROR_AMBIGUOUS_FILENAME) {
@@ -1962,7 +1977,7 @@ int  cm_BPlusDirDeleteEntry(cm_dirOp_t * op, clientchar_t *centry)
             }
         }
     }
-    
+
     QueryPerformanceCounter(&end);
 
     bplus_remove_time += (end.QuadPart - start.QuadPart);
@@ -1972,10 +1987,10 @@ int  cm_BPlusDirDeleteEntry(cm_dirOp_t * op, clientchar_t *centry)
         free(normalizedEntry);
 
     return rc;
-      
+
 }
 
-static 
+
 int cm_BPlusDirFoo(struct cm_scache *scp, struct cm_dirEntry *dep,
                    void *dummy, osi_hyper_t *entryOffsetp)
 {
@@ -1983,7 +1998,7 @@ int cm_BPlusDirFoo(struct cm_scache *scp, struct cm_dirEntry *dep,
     dataT  data;
     normchar_t *normalized_name=NULL;
 
-    cm_SetFid(&data.fid, scp->fid.cell, scp->fid.volume, 
+    cm_SetFid(&data.fid, scp->fid.cell, scp->fid.volume,
               ntohl(dep->fid.vnode), ntohl(dep->fid.unique));
     data.cname = NULL;
     data.fsname = NULL;
@@ -1998,11 +2013,17 @@ int cm_BPlusDirFoo(struct cm_scache *scp, struct cm_dirEntry *dep,
 #endif
         return 0;
     }
-  
+
     data.cname = cm_FsStringToClientStringAlloc(dep->name, -1, NULL);
+    if (data.cname == NULL) {
+#ifdef DEBUG
+        DebugBreak();
+#endif
+        return 0;
+    }
     data.fsname = cm_FsStrDup(dep->name);
     data.shortform = FALSE;
-     
+
     /* the Write lock is held in cm_BPlusDirBuildTree() */
     insert(scp->dirBplus, key, data);
 
@@ -2017,10 +2038,12 @@ int cm_BPlusDirFoo(struct cm_scache *scp, struct cm_dirEntry *dep,
 
         key.name = wshortName;
         data.cname = cm_FsStringToClientStringAlloc(dep->name, -1, NULL);
-        data.fsname = cm_FsStrDup(dep->name);
-        data.shortform = TRUE;
-  
-        insert(scp->dirBplus, key, data);
+        if (data.cname) {
+            data.fsname = cm_FsStrDup(dep->name);
+            data.shortform = TRUE;
+
+            insert(scp->dirBplus, key, data);
+        }
     }
 
     if (normalized_name)
@@ -2129,30 +2152,33 @@ void cm_BPlusDumpStats(void)
     afsi_log("             Free: %-16I64d", bplus_free_time);
 }
 
-static cm_direnum_t * 
+static cm_direnum_t *
 cm_BPlusEnumAlloc(afs_uint32 entries)
 {
     cm_direnum_t * enump;
     size_t	   size;
 
     if (entries == 0)
-	return NULL;
-
-    size = sizeof(cm_direnum_t)+(entries-1)*sizeof(cm_direnum_entry_t);
+        size = sizeof(cm_direnum_t);
+    else
+        size = sizeof(cm_direnum_t)+(entries-1)*sizeof(cm_direnum_entry_t);
     enump = (cm_direnum_t *)malloc(size);
-    memset(enump, 0, size);
-    enump->count = entries;
+    if (enump) {
+        memset(enump, 0, size);
+        enump->count = entries;
+    }
     return enump;
 }
 
-long 
-cm_BPlusDirEnumerate(cm_scache_t *scp, afs_uint32 locked, 
-                     clientchar_t * maskp, cm_direnum_t **enumpp)
+long
+cm_BPlusDirEnumerate(cm_scache_t *dscp, cm_user_t *userp, cm_req_t *reqp,
+                     afs_uint32 locked, clientchar_t * maskp,
+                     afs_uint32 fetchStatus, cm_direnum_t **enumpp)
 {
     afs_uint32 count = 0, slot, numentries;
     Nptr leafNode = NONODE, nextLeafNode;
     Nptr firstDataNode, dataNode, nextDataNode;
-    cm_direnum_t * enump;
+    cm_direnum_t * enump = NULL;
     long rc = 0;
     char buffer[512];
 
@@ -2160,15 +2186,23 @@ cm_BPlusDirEnumerate(cm_scache_t *scp, afs_uint32 locked,
 
     /* Read lock the bplus tree so the data can't change */
     if (!locked)
-	lock_ObtainRead(&scp->dirlock);
+	lock_ObtainRead(&dscp->dirlock);
 
-    if (scp->dirBplus == NULL) {
+    /*
+     * Hold a reference to the directory so that it won't be
+     * recycled while the enumeration is active.
+     */
+    cm_HoldSCache(dscp);
+    cm_HoldUser(userp);
+
+    if (dscp->dirBplus == NULL) {
 	osi_Log0(afsd_logp, "cm_BPlusDirEnumerate No BPlus Tree");
+        rc = CM_ERROR_WOULDBLOCK;
 	goto done;
     }
 
     /* Compute the number of entries */
-    for (count = 0, leafNode = getleaf(scp->dirBplus); leafNode; leafNode = nextLeafNode) {
+    for (count = 0, leafNode = getleaf(dscp->dirBplus); leafNode; leafNode = nextLeafNode) {
 
 	for ( slot = 1, numentries = numentries(leafNode); slot <= numentries; slot++) {
 	    firstDataNode = getnode(leafNode, slot);
@@ -2206,7 +2240,7 @@ cm_BPlusDirEnumerate(cm_scache_t *scp, afs_uint32 locked,
     }
 
     /* Copy the name and fid for each cname entry into the enumeration */
-    for (count = 0, leafNode = getleaf(scp->dirBplus); leafNode; leafNode = nextLeafNode) {
+    for (count = 0, leafNode = getleaf(dscp->dirBplus); leafNode; leafNode = nextLeafNode) {
 
 	for ( slot = 1, numentries = numentries(leafNode); slot <= numentries; slot++) {
 	    firstDataNode = getnode(leafNode, slot);
@@ -2258,16 +2292,32 @@ cm_BPlusDirEnumerate(cm_scache_t *scp, afs_uint32 locked,
 	}
 
 	nextLeafNode = getnextnode(leafNode);
-    }   
+    }
+
+    enump->dscp = dscp;
+    enump->userp = userp;
+    enump->reqFlags = reqp->flags;
+    enump->fetchStatus = fetchStatus;
 
   done:
     if (!locked)
-	lock_ReleaseRead(&scp->dirlock);
+	lock_ReleaseRead(&dscp->dirlock);
 
     /* if we failed, cleanup any mess */
     if (rc != 0) {
 	osi_Log0(afsd_logp, "cm_BPlusDirEnumerate rc != 0");
-	if (enump) {
+
+        /*
+         * release the directory because we failed to generate an enumeration object.
+         * adjust the directory position in the queue to ensure it doesn't get pushed
+         * out by the allocation of a large number of cm_scache objects.
+         */
+        lock_ObtainWrite(&cm_scacheLock);
+        cm_AdjustScacheLRU(dscp);
+        cm_ReleaseSCacheNoLock(dscp);
+        lock_ReleaseWrite(&cm_scacheLock);
+        cm_ReleaseUser(userp);
+        if (enump) {
 	    for ( count = 0; count < enump->count && enump->entry[count].name; count++ ) {
 		free(enump->entry[count].name);
 	    }
@@ -2281,17 +2331,25 @@ cm_BPlusDirEnumerate(cm_scache_t *scp, afs_uint32 locked,
     return rc;
 }
 
-long 
-cm_BPlusDirEnumBulkStat(cm_scache_t *dscp, cm_direnum_t *enump, cm_user_t *userp, cm_req_t *reqp)
+long
+cm_BPlusDirEnumBulkStat(cm_direnum_t *enump)
 {
+    cm_scache_t *dscp = enump->dscp;
+    cm_user_t   *userp = enump->userp;
     cm_bulkStat_t *bsp;
     afs_uint32 count;
-    afs_uint32 code;
+    afs_uint32 code = 0;
+    cm_req_t req;
+
+    cm_InitReq(&req);
+    req.flags = enump->reqFlags;
 
     if ( dscp->fid.cell == AFS_FAKE_ROOT_CELL_ID )
         return 0;
 
     bsp = malloc(sizeof(cm_bulkStat_t));
+    if (!bsp)
+        return ENOMEM;
     memset(bsp, 0, sizeof(cm_bulkStat_t));
 
     for ( count = 0; count < enump->count; count++ ) {
@@ -2308,6 +2366,7 @@ cm_BPlusDirEnumBulkStat(cm_scache_t *dscp, cm_direnum_t *enump, cm_user_t *userp
                      */
                     lock_ReleaseWrite(&tscp->rw);
                     cm_ReleaseSCache(tscp);
+                    enump->entry[count].flags |= CM_DIRENUM_FLAG_GOT_STATUS;
                     continue;
                 }
                 lock_ReleaseWrite(&tscp->rw);
@@ -2319,23 +2378,86 @@ cm_BPlusDirEnumBulkStat(cm_scache_t *dscp, cm_direnum_t *enump, cm_user_t *userp
         bsp->fids[i].Volume = enump->entry[count].fid.volume;
         bsp->fids[i].Vnode = enump->entry[count].fid.vnode;
         bsp->fids[i].Unique = enump->entry[count].fid.unique;
+        enump->entry[count].flags |= CM_DIRENUM_FLAG_GOT_STATUS;
 
         if (bsp->counter == AFSCBMAX) {
-            code = cm_TryBulkStatRPC(dscp, bsp, userp, reqp);
+            code = cm_TryBulkStatRPC(dscp, bsp, userp, &req);
             memset(bsp, 0, sizeof(cm_bulkStat_t));
         }
     }
 
     if (bsp->counter > 0)
-        code = cm_TryBulkStatRPC(dscp, bsp, userp, reqp);
+        code = cm_TryBulkStatRPC(dscp, bsp, userp, &req);
 
     free(bsp);
-    return 0;
+    return code;
 }
 
-long 
+static long
+cm_BPlusDirEnumBulkStatNext(cm_direnum_t *enump)
+{
+    cm_scache_t *dscp = enump->dscp;
+    cm_user_t   *userp = enump->userp;
+    cm_bulkStat_t *bsp;
+    afs_uint32 count;
+    afs_uint32 code = 0;
+    cm_req_t req;
+
+    cm_InitReq(&req);
+    req.flags = enump->reqFlags;
+
+    if ( dscp->fid.cell == AFS_FAKE_ROOT_CELL_ID )
+        return 0;
+
+    bsp = malloc(sizeof(cm_bulkStat_t));
+    if (!bsp)
+        return ENOMEM;
+    memset(bsp, 0, sizeof(cm_bulkStat_t));
+
+    for ( count = enump->next; count < enump->count; count++ ) {
+        cm_scache_t   *tscp = cm_FindSCache(&enump->entry[count].fid);
+        int i;
+
+        if (tscp) {
+            if (lock_TryWrite(&tscp->rw)) {
+                /* we have an entry that we can look at */
+                if (!(tscp->flags & CM_SCACHEFLAG_EACCESS) && cm_HaveCallback(tscp)) {
+                    /* we have a callback on it.  Don't bother
+                     * fetching this stat entry, since we're happy
+                     * with the info we have.
+                     */
+                    lock_ReleaseWrite(&tscp->rw);
+                    cm_ReleaseSCache(tscp);
+                    enump->entry[count].flags |= CM_DIRENUM_FLAG_GOT_STATUS;
+                    continue;
+                }
+                lock_ReleaseWrite(&tscp->rw);
+            }	/* got lock */
+            cm_ReleaseSCache(tscp);
+        }	/* found entry */
+
+        i = bsp->counter++;
+        bsp->fids[i].Volume = enump->entry[count].fid.volume;
+        bsp->fids[i].Vnode = enump->entry[count].fid.vnode;
+        bsp->fids[i].Unique = enump->entry[count].fid.unique;
+        enump->entry[count].flags |= CM_DIRENUM_FLAG_GOT_STATUS;
+
+        if (bsp->counter == AFSCBMAX) {
+            code = cm_TryBulkStatRPC(dscp, bsp, userp, &req);
+            break;
+        }
+    }
+
+    if (bsp->counter > 0 && bsp->counter < AFSCBMAX)
+        code = cm_TryBulkStatRPC(dscp, bsp, userp, &req);
+
+    free(bsp);
+    return code;
+}
+
+long
 cm_BPlusDirNextEnumEntry(cm_direnum_t *enump, cm_direnum_entry_t **entrypp)
-{	
+{
     if (enump == NULL || entrypp == NULL || enump->next >= enump->count) {
 	if (entrypp)
 	    *entrypp = NULL;
@@ -2343,6 +2465,10 @@ cm_BPlusDirNextEnumEntry(cm_direnum_t *enump, cm_direnum_entry_t **entrypp)
 	return CM_ERROR_INVAL;
     }
 
+    if (enump->fetchStatus &&
+        !(enump->entry[enump->next].flags & CM_DIRENUM_FLAG_GOT_STATUS))
+        cm_BPlusDirEnumBulkStatNext(enump);
+
     *entrypp = &enump->entry[enump->next++];
     if ( enump->next == enump->count ) {
 	osi_Log0(afsd_logp, "cm_BPlusDirNextEnumEntry STOPNOW");
@@ -2354,7 +2480,32 @@ cm_BPlusDirNextEnumEntry(cm_direnum_t *enump, cm_direnum_entry_t **entrypp)
     }
 }
 
-long 
+long
+cm_BPlusDirPeekNextEnumEntry(cm_direnum_t *enump, cm_direnum_entry_t **entrypp)
+{
+    if (enump == NULL || entrypp == NULL || enump->next >= enump->count) {
+	if (entrypp)
+	    *entrypp = NULL;
+	osi_Log0(afsd_logp, "cm_BPlusDirPeekNextEnumEntry invalid input");
+	return CM_ERROR_INVAL;
+    }
+
+    if (enump->fetchStatus &&
+        !(enump->entry[enump->next].flags & CM_DIRENUM_FLAG_GOT_STATUS))
+        cm_BPlusDirEnumBulkStatNext(enump);
+
+    *entrypp = &enump->entry[enump->next];
+    if ( enump->next == enump->count ) {
+	osi_Log0(afsd_logp, "cm_BPlusDirPeekNextEnumEntry STOPNOW");
+	return CM_ERROR_STOPNOW;
+    }
+    else {
+	osi_Log0(afsd_logp, "cm_BPlusDirPeekNextEnumEntry SUCCESS");
+	return 0;
+    }
+}
+
+long
 cm_BPlusDirFreeEnumeration(cm_direnum_t *enump)
 {
     afs_uint32 count;
@@ -2362,6 +2513,18 @@ cm_BPlusDirFreeEnumeration(cm_direnum_t *enump)
     osi_Log0(afsd_logp, "cm_BPlusDirFreeEnumeration");
 
     if (enump) {
+        /*
+         * Release the directory object but first adjust its position
+         * in the LRU queue to ensure that it does not get stuck at the
+         * end due to the allocation of a large number of cm_scache
+         * entries in the directory.
+         */
+        lock_ObtainWrite(&cm_scacheLock);
+        cm_AdjustScacheLRU(enump->dscp);
+        cm_ReleaseSCacheNoLock(enump->dscp);
+        lock_ReleaseWrite(&cm_scacheLock);
+        cm_ReleaseUser(enump->userp);
+
 	for ( count = 0; count < enump->count && enump->entry[count].name; count++ ) {
 	    free(enump->entry[count].name);
 	}
@@ -2371,15 +2534,16 @@ cm_BPlusDirFreeEnumeration(cm_direnum_t *enump)
 }
 
 long
-cm_BPlusDirEnumTest(cm_scache_t * dscp, afs_uint32 locked)
+cm_BPlusDirEnumTest(cm_scache_t * dscp, cm_user_t *userp, cm_req_t *reqp, afs_uint32 locked)
 {
     cm_direnum_t * 	 enump = NULL;
     cm_direnum_entry_t * entryp;
     long 	   	 code;
 
+
     osi_Log0(afsd_logp, "cm_BPlusDirEnumTest start");
 
-    for (code = cm_BPlusDirEnumerate(dscp, locked, NULL, &enump); code == 0; ) {
+    for (code = cm_BPlusDirEnumerate(dscp, userp, reqp, locked, NULL, 1, &enump); code == 0; ) {
 	code = cm_BPlusDirNextEnumEntry(enump, &entryp);
 	if (code == 0 || code == CM_ERROR_STOPNOW) {
 	    char buffer[1024];
@@ -2421,7 +2585,7 @@ cm_BPlusDirEnumTest(cm_scache_t * dscp, afs_uint32 locked)
 		    entryp->name,
 		    entryp->fid.cell, entryp->fid.volume, entryp->fid.vnode, entryp->fid.unique,
 		    entryp->shortName,
-		    type, 
+		    type,
 		    dv);
 
 	    osi_Log0(afsd_logp, osi_LogSaveString(afsd_logp, buffer));
