@@ -1,14 +1,14 @@
 /*
  * Copyright 2000, International Business Machines Corporation and others.
  * All Rights Reserved.
- * 
+ *
  * This software has been released under the terms of the IBM Public
  * License.  For details, see the LICENSE file in the top-level source
  * directory or online at http://www.openafs.org/dl/license10.html
  */
 
-#ifndef __CM_USER_H_ENV__
-#define __CM_USER_H_ENV__ 1
+#ifndef OPENAFS_WINNT_AFSD_CM_USER_H
+#define OPENAFS_WINNT_AFSD_CM_USER_H 1
 
 #include <osi.h>
 #include <rx/rxkad.h>
@@ -16,7 +16,7 @@
 /* user structure
  * no free references outside of cm_allUsersp
  * there are held references from cm_conn_t.
- * 
+ *
  * All the fields in this structure are locked by the
  * corresponding userp's userp->mx mutex.
  */
@@ -37,10 +37,11 @@ typedef struct cm_ucell {
 #endif
 } cm_ucell_t;
 
-#define CM_UCELLFLAG_HASTIX	1	/* has Kerberos tickets */
-#define CM_UCELLFLAG_RXKAD	2	/* an rxkad connection */
-#define CM_UCELLFLAG_BADTIX	4	/* tickets are bad or expired */
-#define CM_UCELLFLAG_RXGK       8       /* an rxgk connection */
+#define CM_UCELLFLAG_HASTIX	  1	/* has Kerberos tickets */
+#define CM_UCELLFLAG_RXKAD	  2	/* an rxkad connection */
+#define CM_UCELLFLAG_BADTIX	  4	/* tickets are bad or expired */
+#define CM_UCELLFLAG_RXGK         8     /* an rxgk connection */
+#define CM_UCELLFLAG_ROOTUSER  0x10     /* cm_rootUser connection */
 
 typedef struct cm_user {
     unsigned long refCount;             /* ref count - cm_userLock */
@@ -72,4 +73,4 @@ extern void cm_CheckTokenCache(time_t now);
 
 extern cm_user_t *cm_rootUserp;
 
-#endif /*  __CM_USER_H_ENV__ */
+#endif /*  OPENAFS_WINNT_AFSD_CM_USER_H */
