@@ -1,7 +1,7 @@
 /*
  * Copyright 2000, International Business Machines Corporation and others.
  * All Rights Reserved.
- * 
+ *
  * This software has been released under the terms of the IBM Public
  * License.  For details, see the LICENSE file in the top-level source
  * directory or online at http://www.openafs.org/dl/license10.html
@@ -32,15 +32,15 @@
 
 
 /* Imported variables */
-extern int afs_shuttingdown;
+extern enum afs_shutdown_state afs_shuttingdown;
 
 /* Exported variables */
 afs_uint32 pag_epoch;
-#if defined(UKERNEL) && defined(AFS_WEB_ENHANCEMENTS)
+#if defined(UKERNEL)
 afs_uint32 pagCounter = 1;
 #else
 afs_uint32 pagCounter = 0;
-#endif /* UKERNEL && AFS_WEB_ENHANCEMENTS */
+#endif /* UKERNEL */
 
 #ifdef AFS_LINUX26_ONEGROUP_ENV
 #define NUMPAGGROUPS 1
@@ -73,7 +73,7 @@ afs_uint32 pagCounter = 0;
  * anyway, so the pag is an alternative handle which is somewhat more
  * secure (although of course not absolutely secure).
 */
-#if !defined(UKERNEL) || !defined(AFS_WEB_ENHANCEMENTS)
+#if !defined(UKERNEL)
 afs_uint32
 genpag(void)
 {
@@ -125,7 +125,7 @@ getpag(void)
     return (pagCounter);
 #endif
 }
-#endif /* UKERNEL && AFS_WEB_ENHANCEMENTS */
+#endif /* UKERNEL */
 
 /* used to require 10 seconds between each setpag to guarantee that
  * PAGs never wrap - which would be a security hole.  If we presume
@@ -191,6 +191,8 @@ afs_pag_wait(afs_ucred_t **acred)
 int
 #if	defined(AFS_SUN5_ENV)
 afs_setpag(afs_ucred_t **credpp)
+#elif	defined(AFS_FBSD_ENV)
+afs_setpag(struct thread *td, void *args)
 #elif  defined(AFS_DARWIN_ENV) || defined(AFS_XBSD_ENV)
 afs_setpag(afs_proc_t *p, void *args, int *retval)
 #else
@@ -220,6 +222,11 @@ afs_setpag(void)
 
 #if	defined(AFS_SUN5_ENV)
     code = AddPag(genpag(), credpp);
+#elif	defined(AFS_FBSD_ENV)
+    code = AddPag(td, genpag(), &td->td_ucred);
+#elif   defined(AFS_NBSD40_ENV)
+    /* XXXX won't work */
+    code = AddPag(p, genpag(), (afs_ucred_t **) osi_curcred());
 #elif	defined(AFS_XBSD_ENV)
     code = AddPag(p, genpag(), &p->p_rcred);
 #elif	defined(AFS_AIX41_ENV)
@@ -284,7 +291,7 @@ afs_setpag(void)
     return (code);
 }
 
-#if defined(UKERNEL) && defined(AFS_WEB_ENHANCEMENTS)
+#if defined(UKERNEL)
 /*
  * afs_setpag_val
  * This function is like setpag but sets the current thread's pag id to a
@@ -295,6 +302,8 @@ afs_setpag(void)
 int
 #if	defined(AFS_SUN5_ENV)
 afs_setpag_val(afs_ucred_t **credpp, int pagval)
+#elif	defined(AFS_FBSD_ENV)
+afs_setpag_val(struct thread *td, void *args, int pagval)
 #elif  defined(AFS_DARWIN_ENV) || defined(AFS_XBSD_ENV)
 afs_setpag_val(afs_proc_t *p, void *args, int *retval, int pagval)
 #else
@@ -323,6 +332,8 @@ afs_setpag_val(int pagval)
 
 #if	defined(AFS_SUN5_ENV)
     code = AddPag(pagval, credpp);
+#elif	defined(AFS_FBSD_ENV)
+    code = AddPag(td, pagval, &td->td_ucred);
 #elif	defined(AFS_XBSD_ENV)
     code = AddPag(p, pagval, &p->p_rcred);
 #elif	defined(AFS_AIX41_ENV)
@@ -403,15 +414,16 @@ afs_getpag_val(void)
     return pagvalue;
 }
 #endif
-#endif /* UKERNEL && AFS_WEB_ENHANCEMENTS */
+#endif /* UKERNEL */
 
 
 /* Note - needs to be available on AIX, others can be static - rework this */
-#if defined(AFS_DARWIN_ENV) || defined(AFS_XBSD_ENV)
 int
+#if defined(AFS_FBSD_ENV)
+AddPag(struct thread *p, afs_int32 aval, afs_ucred_t **credpp)
+#elif defined(AFS_DARWIN_ENV) || defined(AFS_XBSD_ENV)
 AddPag(afs_proc_t *p, afs_int32 aval, afs_ucred_t **credpp)
 #else
-int
 AddPag(afs_int32 aval, afs_ucred_t **credpp)
 #endif
 {
@@ -421,6 +433,8 @@ AddPag(afs_int32 aval, afs_ucred_t **credpp)
     AFS_STATCNT(AddPag);
 #if defined(AFS_DARWIN_ENV) || defined(AFS_XBSD_ENV)
     if ((code = setpag(p, credpp, aval, &newpag, 0)))
+#elif defined(AFS_NBSD40_ENV)
+    if ((code = setpag(p, (void *) credpp, aval, &newpag, 0)))
 #else
     if ((code = setpag(credpp, aval, &newpag, 0)))
 #endif
@@ -434,7 +448,7 @@ AddPag(afs_int32 aval, afs_ucred_t **credpp)
 
 
 int
-afs_InitReq(register struct vrequest *av, afs_ucred_t *acred)
+afs_InitReq(struct vrequest *av, afs_ucred_t *acred)
 {
 #if defined(AFS_LINUX26_ENV) && !defined(AFS_NONFSTRANS)
     int code;
@@ -442,7 +456,7 @@ afs_InitReq(register struct vrequest *av, afs_ucred_t *acred)
 
     AFS_STATCNT(afs_InitReq);
     memset(av, 0, sizeof(*av));
-    if (afs_shuttingdown)
+    if (afs_shuttingdown == AFS_SHUTDOWN)
 	return EIO;
 
 #ifdef AFS_LINUX26_ENV
@@ -467,12 +481,67 @@ afs_InitReq(register struct vrequest *av, afs_ucred_t *acred)
 #elif defined(AFS_SUN510_ENV)
         av->uid = crgetruid(acred);
 #else
-	av->uid = afs_cr_uid(acred);	/* default when no pag is set */
+	av->uid = afs_cr_ruid(acred);	/* default when no pag is set */
 #endif
     }
     return 0;
 }
 
+/*!
+ * Allocate and setup a vrequest.
+ *
+ * \note The caller must free the allocated vrequest with
+ *       afs_DestroyReq() if this function returns successfully (zero).
+ *
+ * \note The GLOCK must be held on platforms which require the GLOCK
+ *       for osi_AllocSmallSpace() and osi_FreeSmallSpace().
+ *
+ * \param[out] avpp   address of the vrequest pointer
+ * \param[in]  acred  user credentials to setup the vrequest
+ *                    afs_osi_credp should be used for anonymous connections
+ * \return     0 on success
+ */
+int
+afs_CreateReq(struct vrequest **avpp, afs_ucred_t *acred)
+{
+    int code;
+    struct vrequest *treq = NULL;
+
+    if (afs_shuttingdown == AFS_SHUTDOWN) {
+	return EIO;
+    }
+    if (!avpp || !acred) {
+	return EINVAL;
+    }
+    treq = osi_AllocSmallSpace(sizeof(struct vrequest));
+    if (!treq) {
+	return ENOMEM;
+    }
+    code = afs_InitReq(treq, acred);
+    if (code != 0) {
+	osi_FreeSmallSpace(treq);
+	return code;
+    }
+    *avpp = treq;
+    return 0;
+}
+
+/*!
+ * Deallocate a vrequest.
+ *
+ * \note The GLOCK must be held on platforms which require the GLOCK
+ *       for osi_FreeSmallSpace().
+ *
+ * \param[in]  av  pointer to the vrequest to free; may be NULL
+ */
+void
+afs_DestroyReq(struct vrequest *av)
+{
+    if (av) {
+	osi_FreeSmallSpace(av);
+    }
+}
+
 #ifndef AFS_LINUX26_ONEGROUP_ENV
 afs_uint32
 afs_get_pag_from_groups(gid_t g0a, gid_t g1a)
@@ -490,13 +559,13 @@ afs_get_pag_from_groups(gid_t g0a, gid_t g1a)
 	h = (g0 >> 14);
 	h = (g1 >> 14) + h + h + h;
 	ret = ((h << 28) | l);
-# if defined(UKERNEL) && defined(AFS_WEB_ENHANCEMENTS)
+# if defined(UKERNEL)
 	return ret;
 # else
 	/* Additional testing */
 	if (((ret >> 24) & 0xff) == 'A')
 	    return ret;
-# endif /* UKERNEL && AFS_WEB_ENHANCEMENTS */
+# endif /* UKERNEL */
     }
     return NOPAG;
 }
@@ -509,9 +578,9 @@ afs_get_groups_from_pag(afs_uint32 pag, gid_t * g0p, gid_t * g1p)
     AFS_STATCNT(afs_get_groups_from_pag);
     *g0p = pag;
     *g1p = 0;
-# if !defined(UKERNEL) || !defined(AFS_WEB_ENHANCEMENTS)
+# if !defined(UKERNEL)
     pag &= 0x7fffffff;
-# endif /* UKERNEL && AFS_WEB_ENHANCEMENTS */
+# endif /* UKERNEL */
     g0 = 0x3fff & (pag >> 14);
     g1 = 0x3fff & pag;
     g0 |= ((pag >> 28) / 3) << 14;
@@ -528,7 +597,7 @@ void afs_get_groups_from_pag(afs_uint32 pag, gid_t *g0p, gid_t *g1p)
 }
 #endif
 
-#ifndef AFS_LINUX26_ENV
+#if !defined(AFS_LINUX26_ENV) && !defined(AFS_DARWIN110_ENV)
 static afs_int32
 osi_get_group_pag(afs_ucred_t *cred)
 {
@@ -543,7 +612,15 @@ osi_get_group_pag(afs_ucred_t *cred)
     gids = crgetgroups(cred);
     ngroups = crgetngroups(cred);
 #endif
-#if defined(AFS_DARWIN_ENV) || defined(AFS_XBSD_ENV)
+#if defined(AFS_NBSD40_ENV)
+#warning com afs_ucred_t w/magic will not work
+    if (cred == NOCRED || cred == FSCRED)
+      return NOPAG;
+    if (osi_crngroups(cred) < 3)
+      return NOPAG;
+    g0 = osi_crgroupbyid(cred, 0);
+    g1 = osi_crgroupbyid(cred, 1);
+#elif defined(AFS_DARWIN_ENV) || defined(AFS_XBSD_ENV)
     if (cred == NOCRED || cred == FSCRED)
 	return NOPAG;
     if (cred->cr_ngroups < 3)
@@ -593,6 +670,7 @@ PagInCred(afs_ucred_t *cred)
     if (cred == NULL || cred == afs_osi_credp) {
 	return NOPAG;
     }
+#ifndef AFS_DARWIN110_ENV
 #if defined(AFS_LINUX26_ENV) && defined(LINUX_KEYRING_SUPPORT)
     /*
      * If linux keyrings are in use and we carry the session keyring in our credentials
@@ -602,13 +680,17 @@ PagInCred(afs_ucred_t *cred)
      * With keyrings but no kernel credentials, look at groups first and fall back
      * to looking at the keyrings.
      */
-# if !defined(STRUCT_TASK_HAS_CRED)
+# if !defined(STRUCT_TASK_STRUCT_HAS_CRED)
     pag = osi_get_group_pag(cred);
 # endif
     if (pag == NOPAG)
 	pag = osi_get_keyring_pag(cred);
+#elif defined(AFS_AIX51_ENV)
+    if (kcred_getpag(cred, PAG_AFS, &pag) < 0 || pag == 0)
+	pag = NOPAG;
 #else
     pag = osi_get_group_pag(cred);
+#endif
 #endif
     return pag;
 }
